/*
Binary of Babel
Copyright (C) 2010 Segfault Garden

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <string>
#include "audiomanager.h"
#include "demolib.h"
#include <cassert>

using namespace std;

static void
on_pad_added (GstElement *element,
              GstPad     *pad,
              gpointer    data)
{
  GstPad *sinkpad;
  GstElement *decoder = (GstElement *) data;

  /* We can now link this pad with the vorbis-decoder sink pad */
  g_print ("Dynamic pad created, linking demuxer/decoder\n");
  sinkpad = gst_element_get_static_pad(decoder, "sink");
  gst_pad_link(pad, sinkpad);
  gst_object_unref(sinkpad);
}


AudioManager::AudioManager() {
    hasAudioFile = false;
    int zero = 0;
    GstElement *demuxer, *decoder, *conv, *sink;
    gst_init(&zero, NULL);
    pipeline = gst_pipeline_new("audio-player");
    assert(pipeline);
    source   = gst_element_factory_make("filesrc",       "file-source");
    assert(source);
    demuxer  = gst_element_factory_make("oggdemux",      "ogg-demuxer");
    assert(demuxer);
    decoder  = gst_element_factory_make("vorbisdec",     "vorbis-decoder");
    assert(decoder);
    conv     = gst_element_factory_make("audioconvert",  "converter");
    assert(conv);
    sink     = gst_element_factory_make("autoaudiosink", "audio-output");
    assert(sink);

    if (!pipeline || !source || !demuxer || !decoder || !conv || !sink) {
      throw "Gstreamer element(s) could not be created. Exiting.\n";
    }

    bus = gst_pipeline_get_bus(GST_PIPELINE (pipeline));
    gst_bin_add_many(GST_BIN (pipeline),
                      source, demuxer, decoder, conv, sink, NULL);

    gst_element_link(source, demuxer);
    gst_element_link_many(decoder, conv, sink, NULL);
    g_signal_connect(demuxer, "pad-added", G_CALLBACK (on_pad_added), decoder);
    log_print(debugLevel, "Gstreamer pipeline built.\n");
}

AudioManager::~AudioManager() {
    gst_object_unref (bus);
    gst_element_set_state(pipeline, GST_STATE_NULL);
    gst_object_unref(GST_OBJECT (pipeline));
}

void AudioManager::loadFile(const char *fileName) {
    g_object_set(G_OBJECT (source), "location", fileName, NULL);
    gst_element_set_state(pipeline, GST_STATE_PAUSED);
    hasAudioFile = true;
}

void AudioManager::play() {
    if (hasAudioFile)
        gst_element_set_state(pipeline, GST_STATE_PLAYING);
}

void AudioManager::stop() {
    gst_element_set_state(pipeline, GST_STATE_PAUSED);
}

uint32_t AudioManager::getTime() {
    GstFormat fmt = GST_FORMAT_TIME;
    gint64 pos;

    if (gst_element_query_position (pipeline, &fmt, &pos)) {
        uint32_t result = pos/1000000; // Gstreamer measures in nanoseconds.
        return result;
    }
    throw "Failed to query time from Gstreamer.\n";
}

void AudioManager::pumpEvents() {
    while (gst_bus_have_pending(bus)) {
        GstMessage *msg = gst_bus_pop(bus);
        switch (GST_MESSAGE_TYPE (msg)) {

          case GST_MESSAGE_EOS:
            log_print(debugLevel, "End of audio stream.\n");
            break;

          case GST_MESSAGE_ERROR: {
            gchar  *debug;
            GError *error;

            gst_message_parse_error (msg, &error, &debug);
            g_free (debug);

            string msg = "Gstreamer error: ";
            msg += error->message;
            g_error_free (error);
            throw msg;
          }
          default:
            break;
        }
    }
}
