/*
Binary of Babel
Copyright (C) 2010 Segfault Garden

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <string>
#include <math.h>
#include "SDL_image.h"
#include "SDL_ttf.h"
#include "demorunner.h"
#include "demolib.h"
#include "config.h"
#include "scenerunner.h"

using namespace std;


DemoRunner::DemoRunner(int screenW, int screenH, bool fullScreen) : v(screenW, screenH, fullScreen), scene(*this, v) {
}

void DemoRunner::parseInit(const char *initFile) {
    ifstream inFile;
    string token, name, soname, confFile;
    int w, h;

    inFile.open(initFile);

    if (!inFile)
        throw "Could not load initialisation file";

    while (inFile >> token) {
        PluginThread *pt;
        GLuint texid;

        if (token[0] == '#') {
            char temp[1024];
            inFile.getline(temp, 1024);
            continue;
        }

        if (token == "loglevel") {
            string level;
            inFile >> level;
            parseLogLevel(level);
        } else if (token == "song") {
            string songFile;
            inFile >> songFile;
            audio.loadFile(songFile.c_str());
            log_print(debugLevel, "Loaded audio file %s.\n", songFile.c_str());
        } else if (token == "plugin") {
            textureSize ts;
            inFile >> name;
            inFile >> soname;
            inFile >> w;
            inFile >> h;
            inFile >> confFile;

            if (w <= 0)
                throw "Incorrect width";
            if (h <= 0)
                throw "Incorrect height";
            ts.w = w;
            ts.h = h;
            pt = new PluginThread(soname.c_str(), w, h, confFile.c_str());

            pluginThreads.push_back(pt);
            pluginThreadNameMap[name] = pt;

            pt->continueToNextFrame(0);
            texid = v.createTexture(w, h, 4, false, GL_RGBA, pt->getTexture());
            textureMap[name] = texid;
            textureSizeMap[name] = ts;
            // We leave all plugins in such a state that their hasNewTexture() will
            // return true within a bounded time (i.e. the time it takes to render a frame).
            pt->continueToNextFrame(1);

            log_print(debugLevel, "Initialised plugin %s.\n", name.c_str());
        } else {
            string error = "Unknown command: ";
            error += token;
            throw error;
        }
     }
}

void DemoRunner::parseLogLevel(std::string &level) {
    logLevel newLevel;
    if (level == "silence") {
        newLevel = silenceLevel;
    } else if (level == "error") {
        newLevel = errorLevel;
    } else if (level == "message") {
        newLevel = messageLevel;
    } else if (level == "debug") {
        newLevel = debugLevel;
    } else if (level == "trace") {
        newLevel = traceLevel;
    } else {
        throw "Unkown logging level.";
    }
    set_logging_level(newLevel);
    log_print(debugLevel, "Set logging level to %s.\n", level.c_str());
}


void DemoRunner::updatePluginTexture(GLuint texid, PluginThread *pt) {
    v.updateTexture(texid, pt->getTextureWidth(), pt->getTextureHeight(), pt->getTexture());
}

void DemoRunner::parseFonts(const char *initFile) {
    ifstream inFile;
    string fontName, pointSize, textureName, r, g, b, token;
    int size;
    char text[1024];
    inFile.open(initFile);
    TTF_Font *font = NULL;
    SDL_Color fontColor;
    if (!inFile)
        throw "Could not load texture definition file";

    token = readFirstNonCommentToken(inFile);
    while(token != "") {
        if (token == "font") {
            fontName = readFirstNonCommentToken(inFile);
            pointSize = readFirstNonCommentToken(inFile);
            r = readFirstNonCommentToken(inFile);
            g = readFirstNonCommentToken(inFile);
            b = readFirstNonCommentToken(inFile);
            size = atoi(pointSize.c_str());
            fontColor.r = atoi(r.c_str());
            fontColor.g = atoi(g.c_str());
            fontColor.b = atoi(b.c_str());
            log_print(debugLevel, "Read font file %s at size %d.\n", fontName.c_str(), size);
            if (font != NULL)
                TTF_CloseFont(font);
            font = TTF_OpenFont(fontName.c_str(), size);
            if (!font) {
                string msg = "TTF error: ";
                msg += TTF_GetError();
                throw msg;
            }
        } else {
            textureSize ts;
            SDL_Surface *drawn;
            if (!font)
                throw "Font not defined.";
            textureName = token;
            inFile.ignore(1, ' '); // Skip the space.
            inFile.getline(text, 1024);
            drawn = TTF_RenderUTF8_Blended(font, text, fontColor);
            if (!drawn) {
                string msg = "Error while rendering text: ";
                msg += TTF_GetError();
                throw msg;
            }
            ts.w = drawn->w;
            ts.h = drawn->h;
            textureMap[textureName] = surfaceToTexture(drawn, "linear");
            textureSizeMap[textureName] = ts;
            log_print(debugLevel, "Created overlay %s with text \"%s\".\n", textureName.c_str(), text);
            SDL_FreeSurface(drawn);
        }
        token = readFirstNonCommentToken(inFile);
    }
    if (font)
        TTF_CloseFont(font);
}


void DemoRunner::parseTextures(const char *texFile) {
    ifstream inFile;
    string tex, texname, texfile, texfilter;
    int w, h;

    inFile.open(texFile);

    if (!inFile)
        throw "Could not load texture definition file";

    while (inFile >> tex) {
        if (tex[0] == '#') {
            char temp[1024];
            inFile.getline(temp, 1024);
            continue;
        }

        if (tex != "texture") {
            string error = "Expecting texture, got: ";
            error += tex;
            throw error;
        }
        inFile >> texname;
        inFile >> texfile;
        inFile >> texfilter;
        loadTexture(texname, texfile.c_str(), texfilter);
    }
    log_print(debugLevel, "Texture map has %d entries.\n", (int)textureMap.size());
}

void DemoRunner::loadTexture(const string &texname, const char *fname, string texfilter) {
    SDL_Surface *im = IMG_Load(fname);
    textureSize ts;
    if (!im) {
        string msg = "Failed to load image ";
        msg += fname;
        throw msg.c_str();
    }
    ts.w = im->w;
    ts.h = im->h;
    GLuint result = surfaceToTexture(im, texfilter);
    textureMap[texname] = result;
    textureSizeMap[texname] = ts;
    log_print(debugLevel, "Loaded texture %s to name %s with filter %s.\n", fname, texname.c_str(), texfilter.c_str());
    SDL_FreeSurface(im);

}


GLuint DemoRunner::surfaceToTexture(SDL_Surface *im, string texfilter) {
    GLuint pixelFormat;
    GLuint texid;
    bool smooth;

    if (im->format->BitsPerPixel == 24) {
        pixelFormat = GL_RGB;
    } else if (im->format->BitsPerPixel == 32) {
        pixelFormat = GL_RGBA;
    } else {
        char msg[80];
        sprintf(msg, "Can't handle bit depth of %d.", im->format->BitsPerPixel);
        throw msg;
    }

    if (texfilter == "nearest") {
        smooth = true;
    } else if (texfilter == "linear") {
        smooth = false;
    } else {
        throw "Unknown texture filter type.";
    }

    assert(im->pitch == im->w*im->format->BitsPerPixel/8);
    texid = v.createTexture(im->w, im->h, im->format->BitsPerPixel/8, smooth, pixelFormat, (unsigned char*) im->pixels);
    return texid;
}


void DemoRunner::shutdown() {
    for (int i=0; i<pluginThreads.size(); i++)
        pluginThreads[i]->shutdown();
    log_print(debugLevel, "Closed plugins.\n");

    for (tr1::unordered_map<string, GLuint>::iterator it = textureMap.begin(); it != textureMap.end(); it++) {
        GLuint texid = (*it).second;
        v.deleteTexture(texid);
    }
    log_print(debugLevel, "Deallocated OpenGL textures.\n");

    pluginThreads.clear();
    pluginThreadNameMap.clear();
    textureMap.clear();
    textureSizeMap.clear();

    finalizeGraphics();
    log_print(debugLevel, "Shut down SDL.\n");
}

void DemoRunner::renderTest() {
    audio.play();
    scene.slideshowLoop();
    audio.stop();
}


string DemoRunner::readFirstNonCommentToken(ifstream &iFile) {
    string token;
    char temp[1024];

    iFile >> token;
    while (token[0] == '#') {
        iFile.getline(temp, 1024);
        iFile >> token;
    }
    return token;
}

/*
 * Draw those overlays that are "on" at the specified point in time.
 */

void DemoRunner::drawOverlays(vector<basicOverlay> &overlays, Uint32 currentTime) {
    v.switchTo2DMode();
    for (vector<basicOverlay>::const_iterator it = overlays.begin();
            it != overlays.end(); it++) {
        const basicOverlay &o = *it;
        double alpha = 1.0;
        if (currentTime >= o.startTime && currentTime <= o.endTime) {
            double w, x, y, alpha, fraction;
            if (o.w >= 0)
                w = o.w;
            else {
                // Preserve original's aspect ratio.
                textureSize ts = textureSizeMap[o.texName];
                w = o.h*ts.w/ts.h;
            }
            alpha = 1.0;
            x = o.x;
            y = o.y;
            fraction = ((double) (currentTime - o.startTime))/(o.endTime - o.startTime);
            assert(fraction >= 0.0);
            assert(fraction <= 1.0);

            if (o.type == linearType){
                x = (1.0 - fraction)*o.x + fraction*o.x1;
                y = (1.0 - fraction)*o.y + fraction*o.y1;
            } else if (o.type == smoothType) {
                fraction = 1/(1+exp(-(fraction-0.5)*12));
                if (fraction < 0.0)
                    fraction = 0.0;
                if (fraction > 1.0)
                    fraction = 1.0;
                x = (1.0 - fraction)*o.x + fraction*o.x1;
                y = (1.0 - fraction)*o.y + fraction*o.y1;
            } else if (o.type == fadeIn) {
                alpha = fraction;
            } else if (o.type == fadeOut) {
                alpha = 1.0 - fraction;
            }
            v.drawOverlay(textureMap[o.texName], x, y, w, o.h, alpha);
        }
    }
}

void DemoRunner::runDemo(const char *sceneFile) {
    vector<basicScene> scenes;

    log_print(debugLevel, "Running scene test for file %s.\n", sceneFile);

    parseSceneFile(sceneFile, scenes);
    log_print(debugLevel, "Parsed scene file %s. Now have a total of %d scenes.\n", sceneFile, (int)scenes.size());

    zeroTime = getTime();
    log_print(debugLevel, "Starting demo mainloop at time %d.\n", zeroTime);

    try {
        runScenes(scenes);
    } catch (UserExit *e) {
        log_print(debugLevel, "Exiting due to user request at %d.\n", (int) getTime());
        delete e;
        return;
    }
    log_print(debugLevel, "Finished drawing scenes at %d.\n", (int)getTime());
}

void DemoRunner::runScenes(vector<basicScene> &scenes) {
    Uint32 spentTime = 0;

    audio.play();
    for (int i=0; i<scenes.size(); i++) {
        Uint32 sceneZeroTime = zeroTime + spentTime;
        basicScene &s = scenes[i];

        /* Check which scene type to run and do it. Add new
         * scene types here.
         */
        if (s.name == "testscene") {
            scene.runTestScene(sceneZeroTime, s);
        } else if (s.name == "cube") {
            scene.runCubeScene(sceneZeroTime, s);
        } else if (s.name == "intersect") {
            scene.runIntersectScene(sceneZeroTime, s);
        } else if (s.name == "empty") {
            scene.runEmptyScene(sceneZeroTime, s);
        } else if (s.name == "ray") {
            scene.runRayScene(sceneZeroTime, s);
        } else {
            string error = "Tried to run unknown scene named ";
            error += s.name;
            throw error;
        }
        spentTime += s.duration;
        log_print(debugLevel, "Scene number %d shown. Total scene time: %d\n", i, spentTime);
        audio.pumpEvents();
    }
    audio.stop();
}


void DemoRunner::parseSceneFile(const char *sceneFile, vector<basicScene> &scenes) {
    ifstream inFile;
    string token, includeFile;
    inFile.open(sceneFile);
    if (!inFile)
        throw "Could not load scene file.";

    token = readFirstNonCommentToken(inFile);
    while (token != "") { // We have more scenes.
        basicScene news;
        if (token == "scene") {
            scenes.push_back(news);
            parseSingleScene(inFile, scenes.back());
        } else if (token == "include") {
            includeFile = readFirstNonCommentToken(inFile);
            log_print(debugLevel, "Including file %s.\n", includeFile.c_str());
            parseSceneFile(includeFile.c_str(), scenes);
        } else {
            string msg = "Unknown token ";
            msg += token;
            throw msg;
        }
        token = readFirstNonCommentToken(inFile);
    }
}

void DemoRunner::parseSingleScene(std::ifstream &inFile, basicScene &s) {
    // The "scene" token has been read when this function is called.
    string plugin, overlayName, tmp, overlayType;
    const char *plist = "pluginlist";
    const char *olay = "overlays";

    inFile >> s.name;
    inFile >> s.duration;

    tmp = readFirstNonCommentToken(inFile);
    if (tmp != plist)
        throw "Scene definition expected but not found.";

    inFile >> plugin;
    while (plugin != "end") {
        log_print(traceLevel, "Loaded plugin name %s.\n", plugin.c_str());
        if (pluginThreadNameMap.find(plugin) == pluginThreadNameMap.end()) {
            string msg = "Tried to initialize nonexisting plugin ";
            msg += plugin;
            throw msg;
        }
        s.plugins.push_back(plugin);
        inFile >> plugin;
    }
    inFile >> tmp;
    if (tmp != plist) {
        throw "Plugin list ending is malformed.\n";
    }
    log_print(debugLevel, "Loaded %d plugin definitions.\n", (int)s.plugins.size());

    tmp = readFirstNonCommentToken(inFile);
    if (tmp != olay) {
        throw "Overlay definitions expected but not found.\n";
    }
    overlayName = readFirstNonCommentToken(inFile);
    while (overlayName != "end") {
        basicOverlay o;
        o.texName = overlayName;
        if (textureMap.find(o.texName) == textureMap.end()) {
            string err = "Unkown texture for overlay: ";
            err += o.texName;
            throw err;
        }
        inFile >> overlayType;
        if (overlayType == "static") {
            o.type = staticType;
        } else if (overlayType == "linear") {
            o.type = linearType;
        } else if (overlayType == "smooth") {
            o.type = smoothType;
        } else if (overlayType == "fadein") {
            o.type = fadeIn;
        } else if (overlayType == "fadeout") {
            o.type = fadeOut;
        } else {
            string err = "Unknown overlay type: ";
            err += overlayType;
            throw err;
        }
        inFile >> o.x;
        inFile >> o.y;
        inFile >> o.w;
        inFile >> o.h;
        inFile >> o.startTime;
        inFile >> o.endTime;
        if (o.type == linearType || o.type == smoothType) {
            inFile >> o.x1;
            inFile >> o.y1;
        }
        s.overlays.push_back(o);
        assert(o.startTime >= 0);
        assert(o.endTime >= 0);
        if (o.startTime > o.endTime) {
            string err = "Overlay start is later than end.\n";
            printf("X: %d %d\n", o.startTime, o.endTime);
            throw err;
        }
        assert(o.h > 0);
        assert(o.w >= -1); // In this case calculate width from height when rendering.
        assert(o.w != 0);
        overlayName = readFirstNonCommentToken(inFile);
    }
    inFile >> tmp;
    if (tmp != olay) {
        throw "Overlay list ending is malformed.\n";
    }

    tmp = readFirstNonCommentToken(inFile);
    if (tmp != "end") {
        throw "Malformed scene end.\n";
    }
    tmp = readFirstNonCommentToken(inFile);
    if (tmp != "scene") {
        throw "Malformed scene end.\n";
    }

    log_print(debugLevel, "Loaded %d overlays.\n", (int)s.overlays.size());
    log_print(debugLevel, "Loaded scene %s.\n", s.name.c_str());
}

void DemoRunner::initGraphics() {
    v.initGraphics();
    int sdlimageFlags = IMG_INIT_JPG|IMG_INIT_PNG;
    int sdlimageInit = IMG_Init(sdlimageFlags);
    if (sdlimageInit & sdlimageFlags != sdlimageFlags) {
        log_print(errorLevel, "IMG_Init failed: %s\n", IMG_GetError());
        abort();
    }
    if (TTF_Init() != 0) {
        log_print(errorLevel, "SDL ttf init failed: %s\n", TTF_GetError());
        abort();
    }
}

void DemoRunner::finalizeGraphics() {
    TTF_Quit();
    IMG_Quit();
    v.shutdown();
}

Uint32 DemoRunner::getTime() {
    //return audio.getTime();
    return SDL_GetTicks();
}
