/*
Binary of Babel
Copyright (C) 2010 Segfault Garden

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include"SDL.h"
#include"SDL_opengl.h"
#include"SDL_image.h"
#include<stdio.h>
#include<stdlib.h>
#include<math.h>

#include"demoplugin.h"
#include"config.h"

const int WIDTH = 800;
const int HEIGHT = 600;
int texh = 256;
int texw = 256;

SDL_Surface *init_graphics(int w, int h, bool fullscreen) {
    Uint32 surfaceFlags = SDL_OPENGL;
    SDL_Surface *s;
    int depth = 32;

    if(fullscreen)
        surfaceFlags |= SDL_FULLSCREEN;

    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
    s = SDL_SetVideoMode(w, h, depth, surfaceFlags);
    return s;
}

void opengl_setup() {
    glShadeModel(GL_SMOOTH);
    glClearColor(1.0f, 1.0f, 1.0f, 0.0f);
    glClearDepth(1.0f);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    gluPerspective(45.0f,(GLfloat)WIDTH/(GLfloat)HEIGHT,0.1f,100.0f);

    glMatrixMode(GL_MODELVIEW);

}

void simple_draw(double dx, double dy) {
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();
    glTranslated(dx, dy, -3.5f);
    glBegin(GL_QUADS);
    glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f, -1.0f,  1.0f);
    glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f, -1.0f,  1.0f);
    glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f,  1.0f,  1.0f);
    glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f,  1.0f,  1.0f);
    glEnd();

    glFinish();
}

void viewTest(const char *pluginfile, const char *conffile) {
    SDL_Event e;
    Uint32 firstFrameTime;
    DemoPlugin pl(pluginfile);
    pl.load();
    pl.init(texw, texh, conffile);
    glColor3d(1.0, 1.0, 1.0);
    GLuint texid[1];
    Uint32 fpsCountStart;
    int framesRendered = 0;

    glGenTextures(1, &texid[0]);
    glBindTexture(GL_TEXTURE_2D, texid[0]);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

    firstFrameTime = fpsCountStart =  SDL_GetTicks();
    while(1) {
        const Uint32 second = 1000;
        const double frameTimeWith60FPS = second/60.0;
        Uint32 startTime = SDL_GetTicks();
        Uint32 endTime, optimalFrameTime;
        const Uint32 delayPrecision = 10;
        pl.render(startTime - firstFrameTime);
        glTexImage2D(GL_TEXTURE_2D, 0, 4, texw, texh, 0, GL_RGBA, GL_UNSIGNED_BYTE,
                pl.getTexture());
        simple_draw(0.0, 0.0);
        endTime = SDL_GetTicks();

        // At most 60 FPS.
        optimalFrameTime = fpsCountStart + (framesRendered+1)*frameTimeWith60FPS;
        if (optimalFrameTime > endTime) {
            SDL_Delay(optimalFrameTime - endTime);
        }
        SDL_GL_SwapBuffers();
        framesRendered++;

        // Print FPS rate.
        if (SDL_GetTicks() - fpsCountStart > (second - delayPrecision)) {
            printf("FPS: %d\n", framesRendered);
            framesRendered = 0;
            fpsCountStart = SDL_GetTicks();
        }

        // Check if we need to exit.
        while(SDL_PollEvent(&e)) {
            if (e.type == SDL_QUIT ||
                (e.type == SDL_KEYDOWN &&
                (e.key.keysym.sym == SDLK_ESCAPE || e.key.keysym.sym == SDLK_q))) {
                pl.shutdown();
                glDeleteTextures(1, &texid[0]);
                return;
            }
        }
    }
}

int main(int argc, char **argv) {
    const int isFullscreen = false;
    SDL_Surface *canvas;
    const char *conffile;

    if(argc < 2) {
        printf("No plugin specified.\n");
        return 1;
    }
    if(strstr(argv[1], "/") == 0) {
        printf("Slash missing from plugin file name. Exiting as this will most likely fail.\n");
        return 1;
    }

    printf("Simple plugin viewer\n");

    if(SDL_Init(SDL_INIT_AUDIO|SDL_INIT_VIDEO) < 0) {
        fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
        return 1;
    }
    SDL_WM_SetCaption("Plugin viewer", "Plugin viewer");

    int sdlimageFlags = IMG_INIT_JPG|IMG_INIT_PNG;
    int sdlimageInit = IMG_Init(sdlimageFlags);
    if(sdlimageInit & sdlimageInit != sdlimageFlags) {
        printf("IMG_Init failed: %s\n", IMG_GetError());
        return 1;
    }

    canvas = init_graphics(WIDTH, HEIGHT, isFullscreen);
    if(!canvas) {
        fprintf(stderr, "Could not initialize video: %s\n", SDL_GetError());
        exit(1);
    }

    if(argc<2)
        conffile = 0;
    else
        conffile = argv[2];
    if (argc == 5) {
        printf("Using custom texture size.\n");
        texw = atoi(argv[3]);
        texh = atoi(argv[4]);
    }
    opengl_setup();
    try {
        viewTest(argv[1], conffile);
    } catch (char *ex) {
        printf("Got exception: %s\n", ex);
        return 1;
    }

    IMG_Quit();
    SDL_Quit();
    return 0;
}
