/*
Binary of Babel
Copyright (C) 2010 Segfault Garden

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <jni.h>
#include <string.h>
#include <assert.h>

#include "pluginapi.h"

static const char *Config_file;
static int Width, Height, Texture_size;

#define USER_CLASSPATH "javaclasses.jar"
 
static JNIEnv *env;
static JavaVM *jvm;
static jint res;
static jclass cls;
static jmethodID render_method_id;
static jmethodID gc_method_id;

void java_fail_if_null(void *value, char *doing_what) {
    if (value == NULL) {
        fprintf(stderr, "Failure in Java code while %s!\n", doing_what);
        if ((*env)->ExceptionOccurred(env)) {
            (*env)->ExceptionDescribe(env);
        }
        abort();
    }
}

void java_fail_if_exception_occurred() {
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        abort();
    }
}

void init_java() {
    JavaVMInitArgs vm_args;
    JavaVMOption options[1];
    options[0].optionString =
        "-Djava.class.path=" USER_CLASSPATH;
    vm_args.version = 0x00010002;
    vm_args.options = options;
    vm_args.nOptions = 1;
    vm_args.ignoreUnrecognized = JNI_TRUE;
    /* Create the Java VM */
    res = JNI_CreateJavaVM(&jvm, (void**)&env, &vm_args);
 
    if (res < 0) {
        fprintf(stderr, "Can't create Java VM\n");
        abort();
    }

    jmethodID init_method_id;
    jstring jstr;
    java_fail_if_null(cls = (*env)->FindClass(env, "Demo"), "finding class Demo");
    java_fail_if_null(init_method_id = (*env)->GetStaticMethodID(env, cls, "init", "(IILjava/lang/String;)V"), "looking up init method");
    java_fail_if_null(render_method_id = (*env)->GetStaticMethodID(env, cls, "render", "(I)[B"), "looking up render method");    
    java_fail_if_null(gc_method_id = (*env)->GetStaticMethodID(env, cls, "gc", "()V"), "looking up gc method");    
    java_fail_if_null(jstr = (*env)->NewStringUTF(env, Config_file), "creating string for config file name");

    jint width = Width, height = Height;
    (*env)->CallStaticVoidMethod(env, cls, init_method_id, width, height, jstr);

    java_fail_if_exception_occurred();
}

void init(int width, int height, const char *config_file)  {
    assert(config_file);
    Config_file = strdup(config_file);
    Texture_size = width * height * 4;
    Width = width;
    Height = height;

    init_java();
}

void render(unsigned char *texture, uint32_t time_in_millis) {
    jint time = time_in_millis;
    jobject result = (*env)->CallStaticObjectMethod(env, cls, render_method_id, time);
    java_fail_if_exception_occurred();

    jsize array_length = (*env)->GetArrayLength(env, result);
    int expected_array_length = Width * Height * 4;

    if (array_length != expected_array_length) {
        fprintf(stderr, "Expected array of %d elements from Java code, got %d elements.\n",
                expected_array_length, array_length);
        abort();
    }
    (*env)->GetByteArrayRegion(env, result, 0, array_length, (signed char*)texture);
    java_fail_if_exception_occurred();
}

void gc(void) {
    (*env)->CallStaticVoidMethod(env, cls, gc_method_id);
    java_fail_if_exception_occurred();
}

void shutdown(void) {
    fprintf(stderr, "Java plugin: Shutting down.\n");
    fprintf(stderr, "NOT destroying the Java VM since the DestroyJavaVM call doesn't work.\n");
    //(*jvm)->DestroyJavaVM(jvm);
    //java_fail_if_exception_occurred();
}

