/*
Binary of Babel
Copyright (C) 2010 Segfault Garden

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <v8.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <assert.h>
#include <math.h>
#include "SDL.h"
#include "SDL_image.h"
#include "pluginapi.h"

using namespace v8;

Persistent<Context> context;
Handle<Script> script;

static uint32_t timeHolder;
static unsigned char *t;
SDL_Surface *backImage;
static int w, h;
const int MAXVAL = 255;

static int clamp(int var, const int maxVal) {
    if (var < 0)
        return 0;
    if (var > maxVal)
        return maxVal;
    return var;
}

static Handle<Value> copyline(const Arguments &args) {
    int sourceLineNum = args[0]->IntegerValue() % MAXVAL;
    int destLineNum = args[1]->IntegerValue() % MAXVAL;

    if (sourceLineNum < 0)
        sourceLineNum = -sourceLineNum;
    if (destLineNum < 0)
        destLineNum = -destLineNum;

    uint32_t *src = (uint32_t*)backImage->pixels;
    uint32_t *dst = (uint32_t*)t;
    for (int j=0; j<h; j++) {
        int soff = j*w + sourceLineNum;
        int doff = j*w + destLineNum;
        dst[doff] = src[soff];
    }
    return Undefined();
}


static void clearTexture(unsigned char *tex) {
    int i, j;

    for (j=0; j<h; j++) {
        for (i=0; i<w; i++) {
            unsigned char *base = tex + 4*(j*w + i);
            base[0] = 0;
            base[1] = 0;
            base[2] = 0;
        }
    }
}

static Handle<Value> TimeCallback(const Arguments& args) {
    return Integer::NewFromUnsigned(timeHolder);
}

static Handle<Value> p0yCallback(const Arguments& args) {
    return Number::New(1.2*sin(timeHolder/900.0));
}

static Handle<Value> p1yCallback(const Arguments& args) {
    return Number::New(1.2*sin(timeHolder/495.0));
}

static void indexToColor(unsigned char c, unsigned char *target) {
    if (c < 50) {
        // Water
        target[0] = 32;
        target[1] = 63;
        target[2] = 255;
    } else if (c < 100) {
        // Forest
        target[0] = 6;
        target[1] = 159;
        target[2] = 2;
    } else if (c < 150) {
        // Brown
        target[0] = 181;
        target[1] = 121;
        target[2] = 6;
    } else if (c < 200) {
        // Rock
        target[0] = 221;
        target[1] = 212;
        target[2] = 195;
    } else {
        // Snow
        target[0] = 255;
        target[1] = 255;
        target[2] = 255;
    }
}


// Does no bounds checking.
static void setPixel(int i, int j, unsigned char *c) {
    unsigned char *dst = t + (j*w + i)*4;
    dst[0] = c[0];
    dst[1] = c[1];
    dst[2] = c[2];
}

static Handle<Value> DrawCallback(const Arguments& args) {
    if (args.Length() != 5) {
        return v8::ThrowException(v8::String::New("Incorrect number of parameters"));
    }
    double x = args[0]->NumberValue();
    double y = args[1]->NumberValue();
    double wid = args[2]->NumberValue();
    double hei = args[3]->NumberValue();
    double c = args[4]->NumberValue();
    unsigned char pixelColor[3];

    int x0 = clamp(w*(0.5*(x - 0.5*wid + 1.0)), MAXVAL);
    int y0 = clamp(h*(0.5*(y - 0.5*hei + 1.0)), MAXVAL);
    int x1 = clamp(w*(0.5*(x + 0.5*wid + 1.0)), MAXVAL);
    int y1 = clamp(h*(0.5*(y + 0.5*hei + 1.0)), MAXVAL);
    assert(x1 >= x0);
    assert(y1 >= y0);
    unsigned char col = clamp(c*MAXVAL, MAXVAL);

    indexToColor(col, pixelColor);
    for (int j=y0; j<y1; j++) {
        for (int i=x0; i<x1; i++) {
            setPixel(i, j, pixelColor);
        }
    }

    return Undefined();
}


void init(int width, int height, const char *config_file) {
    HandleScope handle_scope;
    std::ifstream is;
    int length;
    char * buffer;

    w = width;
    h = height;
    backImage = IMG_Load("../data/puukko.png");
    if (!backImage) {
        printf("Image load is fail.\n");
        return;
    }

    is.open(config_file, std::ios::binary );
    is.seekg (0, std::ios::end);
    length = is.tellg();
    is.seekg (0, std::ios::beg);
    buffer = new char[length];
    is.read (buffer, length);
    is.close();

    // Build callbacks.

    Handle<ObjectTemplate> global = ObjectTemplate::New();
    global->Set(String::New("getTime"), FunctionTemplate::New(TimeCallback));
    global->Set(String::New("getp0y"), FunctionTemplate::New(p0yCallback));
    global->Set(String::New("getp1y"), FunctionTemplate::New(p1yCallback));
    global->Set(String::New("drawBox"), FunctionTemplate::New(DrawCallback));
    global->Set(String::New("copyLine"), FunctionTemplate::New(copyline));
     context = Context::New(NULL, global);
    Context::Scope context_scope(context);

    // Compile script.

    Handle<String> source = String::New(buffer, length);
    String::New(buffer, length);
    script = Script::Compile(source);
    if (script.IsEmpty()) {
        printf("Failed to compile script.\n");
        exit(1);
    }
    delete[] buffer;
}

void render(unsigned char *texture, uint32_t time_in_milliseconds) {
    //clearTexture(texture);
    t = texture;
    timeHolder = time_in_milliseconds;

    HandleScope handle_scope;
    Context::Scope context_scope(context);
    TryCatch trycatch;
    Handle<Value> v = script->Run();
    if (v.IsEmpty()) {
        Handle<Value> exception = trycatch.Exception();
        String::AsciiValue exception_str(exception);
        printf("Javascript exception: %s\n", *exception_str);
        exit(2);
    }
//    printf("%d\n", v->Int32Value());
}

void gc(void) {

}

void shutdown(void) {
    SDL_FreeSurface(backImage);
    script.Clear();
    context.Dispose();
    V8::Dispose();
}
