/*
Binary of Babel
Copyright (C) 2010 Segfault Garden

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "pluginapi.h"
#include <stdio.h>
#include <math.h>
#include <stdint.h>
#include <string.h>

#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>

static lua_State *L;
static uint32_t *_texture;
static struct {
    int width;
    int height;
    int size;
} dimensions;

static const float invgamma = 1 / 2.2;

/* ---------- The plugin itself ---------- */

int setPixel(lua_State *);
int drawLine(lua_State *);
int clearTexture(lua_State *);

void init(int width, int height, const char *config_file) {
    if (config_file == NULL) {
        fprintf(stderr, "Lua plugin expects script as config file\n");
        return;
    }

    dimensions.width = width;
    dimensions.height = height;
    dimensions.size = width * height;

    /* Init interpreter */    
    L = luaL_newstate();
    luaL_openlibs(L);

    /* Register utility functions */
    lua_register(L, "cleartexture", clearTexture);
    lua_register(L, "setpixel", setPixel);
    lua_register(L, "drawline", drawLine);

    /* Load script and call init() */
    if (0 != luaL_dofile(L, config_file)) {
        fprintf(stderr, "Error loading %s\n", config_file);
        return;
    }
    lua_getglobal(L, "init");
    lua_pushinteger(L, width);
    lua_pushinteger(L, height);
    lua_call(L, 2, 0);
}

void render(unsigned char *texture, uint32_t time_in_milliseconds) {
    _texture = (uint32_t *) texture;

    lua_getglobal(L, "render");
    lua_pushinteger(L, time_in_milliseconds);
    lua_call(L, 1, 0);
}

void gc() {
    lua_gc(L, LUA_GCCOLLECT, 0);
}

void shutdown() {
    lua_close(L);
}
/* ---------- C functions for plugin ---------- */

int_fast32_t clamp (int_fast32_t x) {
    return ((x < 0) ? 0 : ((x > 255) ? 255 : x));
}

void plot(int_fast16_t x, int_fast16_t y, uint8_t r, uint8_t g, uint8_t b) {
    int loc = dimensions.width * y + x;
    _texture[loc] = r | g << 8 | b << 16 | 0xFF << 24;
}

/* Plot with brightness */
void plotwb(int_fast16_t x, int_fast16_t y, 
            uint8_t r, uint8_t g, uint8_t b, 
            float br) {
    float factor = powf(br, invgamma);
    plot(x, y, factor * r, factor * g, factor * b);
}

/* Xiaolin Wu antialiased line drawing algorithm */

#define FPART(x) ((x) - floorf(x))
#define ROUND(x) floorf((x) + 0.5)
#define RFPART(x) (1 - FPART(x))

#define FSWAP(x, y) {float tmp = x; x = y; y = tmp;}

void drawLineRaw(float x0, float y0,
                 float x1, float y1, 
                 int_fast8_t r, int_fast8_t g, int_fast8_t b) {
    float dx = x1 - x0;
    float dy = y1 - y0;
    uint8_t steep = fabsf(dx) < fabsf(dy) ? 1 : 0;

    if (steep) {
        FSWAP(x0, y0)
        FSWAP(x1, y1)
        FSWAP(dx, dy)
    }
    if (x1 < x0) {
        FSWAP(x0, x1)
        FSWAP(y0, y1)
    }
    float gradient = dy / dx;

    // First endpoint
    float xend = ROUND(x0);
    float yend = y0 + gradient * (xend - x0);
    float xgap = RFPART(x1 + 0.5);
    int_fast16_t xpxl0 = xend;
    int_fast16_t ypxl0 = floor(yend);
    if (steep) {
        plotwb(ypxl0, xpxl0, r, g, b, RFPART(yend) * xgap);
        plotwb(ypxl0, xpxl0 + 1, r, g, b, FPART(yend) * xgap);
    } else {
        plotwb(xpxl0, ypxl0, r, g, b, RFPART(yend) * xgap);
        plotwb(xpxl0, ypxl0 + 1, r, g, b, FPART(yend) * xgap);
    }
    float intery = yend + gradient;
    
    // Second endpoint
    xend = round(x1);
    yend = y1 + gradient * (xend - x1);
    xgap = FPART(x1 + 0.5);
    int_fast16_t xpxl1 = xend;  // this will be used in the main loop
    int_fast16_t ypxl1 = floorf(yend);
    /* if (steep) { */
    /*     plotwb(ypxl1, xpxl1, r, g, b, RFPART(yend) * xgap); */
    /*     plotwb(ypxl1, xpxl1 + 1, r, g, b, FPART(yend) * xgap); */
    /* } else { */
    /*     plotwb(xpxl1, ypxl1, r, g, b, RFPART(yend) * xgap); */
    /*     plotwb(xpxl1, ypxl1 + 1, r, g, b, FPART(yend) * xgap); */
    /* } */
    int x;
    for (x = xpxl0 + 1; x < xpxl1; x++) {
        if (steep) {
            plotwb(floorf(intery), x, r, g, b, RFPART(intery));
            plotwb(floorf(intery) + 1, x, r, g, b, FPART(intery));
        } else {
            plotwb(x, floorf(intery), r, g, b, RFPART(intery));
            plotwb(x, floorf(intery) + 1, r, g, b, FPART(intery));
        }
        intery = intery + gradient;
    }

}

int drawLine(lua_State *L) {
    int argc = lua_gettop(L);
    if (argc != 5 && argc != 7) {
        lua_pushstring(L, "incorrect argument count");
        lua_error(L);
    }

    float x0 = luaL_optnumber(L, 1, -1);
    float y0 = luaL_optnumber(L, 2, -1);
    float x1 = luaL_optnumber(L, 3, -1);
    float y1 = luaL_optnumber(L, 4, -1);
    int r, g, b;
    if (argc == 5)
        r = g = b = clamp(luaL_optint(L, 5, -1));
    else {
        r = clamp(luaL_optint(L, 5, -1));
        g = clamp(luaL_optint(L, 6, -1));
        b = clamp(luaL_optint(L, 7, -1));
    }

    drawLineRaw(x0, y0, x1, y1, r, g, b);

    return 0;
}

int setPixel(lua_State *L) {
    int argc = lua_gettop(L);
    if (argc != 3 && argc != 5) {
        lua_pushstring(L, "incorrect argument count");
        lua_error(L);
    }

    int x = luaL_optint(L, 1, -1);
    int y = luaL_optint(L, 2, -1);
    int r, g, b;
    if (argc == 3)
        r = g = b = clamp(luaL_optint(L, 3, -1));
    else {
        r = clamp(luaL_optint(L, 3, -1));
        g = clamp(luaL_optint(L, 4, -1));
        b = clamp(luaL_optint(L, 5, -1));
    }

    plot(x, y, r, g, b);

    return 0;
}

int clearTexture(lua_State *L) {
    int i;
    for (i = 0; i < dimensions.size; i++)
        _texture[i] = 0xFF000000;
    return 0;
}
