/*
Binary of Babel
Copyright (C) 2010 Segfault Garden

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "pluginapi.h"

#include <stdio.h>
#include <unistd.h>

/*
 * This plugin wraps an external process, for those languages that
 * just can not be called from C (think D, its lack of dynamic
 * libraries and the runtime that is started only by static Tango
 * lib).
 *
 * The communication protocol with external process is following:
 *
 * - Program name is given as config_file
 *
 * - Plugin starts the program with texture width and height as
 *   command line arguments.
 *
 * - A frame is requested by writing the requested time as
 *   milliseconds to process' standard input as a 32-bit integer
 *   binary.
 *
 * - The process writes the frame to its standard output as sequence
 *   of 4-character binary RGBA values.
 *
 * - The process' stdin is closed when it is no more needed.
 */

static struct {
    int width;
    int height;
    size_t size;
} dim;

static struct {
    const char *name;
    int stdin_wrt;
    int stdout_rd;
} childproc;

void init(int width, int height, const char *config_file) {
    dim.width = width;
    dim.height = height;
    dim.size = width * height * 4;

    if (config_file == NULL) {
        fprintf(stderr, "processwrapper: expected: command\n");
    }

    /* The fork / execve exercise */
    int pipefd_in[2]; /* Input to child */
    int pipefd_out[2]; /* Output from child */

    pipe(pipefd_in);
    pipe(pipefd_out);

    pid_t fork_res = fork();
    if (fork_res == -1) {
        perror(NULL);
        return;
    }

    if (fork_res == 0) {
        /* Child */
        close(pipefd_in[1]);
        close(pipefd_out[0]);

        dup2(pipefd_in[0], STDIN_FILENO);
        dup2(pipefd_out[1], STDOUT_FILENO);

        char width_str[10];
        sprintf(width_str, "%d", width);
        char height_str[10];
        sprintf(height_str, "%d", height);
        execlp(config_file, config_file, width_str, height_str, (char *)NULL);
    } else {
        /* Parent */
        close(pipefd_in[0]);
        close(pipefd_out[1]);

        childproc.name = config_file;
        childproc.stdin_wrt = pipefd_in[1];
        childproc.stdout_rd = pipefd_out[0];
    }
}


void render(unsigned char *texture, uint32_t time_in_milliseconds) {
    write(childproc.stdin_wrt, &time_in_milliseconds, sizeof(uint32_t));

    size_t pos = 0;
    while (pos < dim.size) {
        int count = read(childproc.stdout_rd, texture + pos, dim.size - pos);
        if (count == 0) {
            fprintf(stderr, "processwrapper: %s closed stdout\n", childproc.name);
            return;
        }
        pos += count;
    }
}

void gc(void) {
    /* Could pass a magic number to child, but not useful so far */
}

void shutdown(void) {
    close(childproc.stdin_wrt);
    close(childproc.stdout_rd);
}
