/*
Binary of Babel
Copyright (C) 2010 Segfault Garden

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <Python.h>
#include "pluginapi.h"
#include <sys/types.h>
#include <sys/stat.h>

static const char *Config_file;
static PyObject *Main_module;
static int Width, Height, Texture_size;
static time_t Config_file_mtime;

static time_t file_mtime(const char *filename)
{
    struct stat statdata;
    if (0 != stat(filename, &statdata)) {
        printf("Error stat'ing config file %s.", filename);
        abort();
    }
    return statdata.st_mtime;
}

static int config_file_changed()
{
    time_t current_mtime = file_mtime(Config_file);
    if (current_mtime != Config_file_mtime) {
        Config_file_mtime = current_mtime;
        return 1;
    } else {
        Config_file_mtime = current_mtime;
        return 0;
    }
}

#define PYTHON_FAIL_IF_NULL(value) if (!(value)) python_fail(#value " == NULL")

static void python_fail(const char *why) {
    fprintf(stderr, "Python fail: %s\n", why);
    PyErr_PrintEx(0);
    abort();
}

static void reload_config() {
    char buf[256];
    snprintf(buf, 256, "print 'Loading Python script %s';execfile('%s');", Config_file, Config_file);
    PyRun_SimpleString(buf);

    Main_module = PyImport_AddModule("__main__");
    PYTHON_FAIL_IF_NULL(Main_module);

    /* Call the init function in the loaded script. */
    PyObject *init_function = PyObject_GetAttrString(Main_module, "init");
    PYTHON_FAIL_IF_NULL(init_function);
    PyObject *args = Py_BuildValue("(ii)", Width, Height);
    PYTHON_FAIL_IF_NULL(args);
    PyObject *return_value = PyObject_Call(init_function, args, NULL);
    PYTHON_FAIL_IF_NULL(return_value);
    Py_DECREF(args);
    Py_DECREF(return_value);
}

void init(int width, int height, const char *config_file)  {
    Py_Initialize();
    Config_file = strdup(config_file);
    Texture_size = width * height * 4;
    Width = width;
    Height = height;

    PyRun_SimpleString("import gc"); // Needed in gc() function, import once here.
    reload_config();
}

void render(unsigned char *texture, uint32_t time_in_millis) {
    if (config_file_changed()) {
        printf("Python: Config file changed, reloading config.\n");
        reload_config();
    }
    PyObject *render_function = PyObject_GetAttrString(Main_module, "render");
    PYTHON_FAIL_IF_NULL(render_function);
    PyObject *args = Py_BuildValue("(L)", time_in_millis);
    PYTHON_FAIL_IF_NULL(args);
    PyObject *return_value = PyObject_Call(render_function, args, NULL);
    PYTHON_FAIL_IF_NULL(return_value);
    Py_DECREF(args);
    PyBufferProcs *procs = return_value->ob_type->tp_as_buffer;
    assert(procs);
    void *buf;
    procs->bf_getreadbuffer(return_value, 0, &buf);
    memcpy(texture, buf, Texture_size);
    Py_DECREF(return_value);
}

void gc(void) {
    PyRun_SimpleString("gc.collect()");
}

/* Do we really need this? The Python interpreter cannot clean after
   itself completely, and this may apply to other languages as
   well. */
void shutdown(void) {
    Py_Finalize();
}
