/*
Binary of Babel
Copyright (C) 2010 Segfault Garden

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <libguile.h>
#include "pluginapi.h"

static const char *Config_file;
static int Width, Height, Texture_size;
static time_t Config_file_mtime;

time_t file_mtime(const char *filename)
{
    struct stat statdata;
    if (0 != stat(filename, &statdata)) {
        printf("Error stat'ing config file %s.", filename);
        abort();
    }
    return statdata.st_mtime;
}

int config_file_changed()
{
    time_t current_mtime = file_mtime(Config_file);
    if (current_mtime != Config_file_mtime) {
        Config_file_mtime = current_mtime;
        return 1;
    } else {
        Config_file_mtime = current_mtime;
        return 0;
    }
}

void fail_while(char *doing_what) {
    fprintf(stderr, "Failure in Scheme code while %s!\n", doing_what);
    abort();
}

void reload_config() {
    SCM initFunc;
    scm_c_primitive_load(Config_file);
    Config_file_mtime = file_mtime(Config_file);
    initFunc = scm_variable_ref(scm_c_lookup("init"));
    scm_call_2(initFunc, scm_int2num(Width), scm_int2num(Height));
}

void init_scheme(void) {
    scm_init_guile();
    reload_config();
}

void init(int width, int height, const char *config_file)  {
    Config_file = strdup(config_file);
    Texture_size = width * height * 4;
    Width = width;
    Height = height;
    init_scheme();
}

void render(unsigned char *texture, uint32_t time_in_millis) {
    if (config_file_changed()) {
        printf("scheme: Config file changed, reloading config.\n");
        reload_config();
    }

    SCM renderFunc = scm_variable_ref(scm_c_lookup("render"));
    SCM renderResult = scm_call_1(renderFunc, scm_int2num(time_in_millis));

    scm_t_array_handle arrayHandle;
    scm_array_get_handle(renderResult, &arrayHandle);
    if (scm_array_handle_rank(&arrayHandle) != 2) 
        fail_while("reading render result: wrong array rank");
    const scm_t_array_dim *arrayDimensions = scm_array_handle_dims(&arrayHandle);
    int resultWidth = arrayDimensions[1].ubnd - arrayDimensions[1].lbnd + 1;
    int resultHeight = arrayDimensions[0].ubnd - arrayDimensions[0].lbnd + 1;

    if (resultWidth != Width || resultHeight != Height) {
        fprintf(stderr, "Expected %d x %d array, got %d x %d array.\n", 
                Width, Height, resultWidth, resultHeight);
        fail_while("reading render result: wrong array size");        
    }
    const scm_t_uint32 *pixels = scm_array_handle_u32_elements(&arrayHandle);
    memcpy(texture, pixels, Texture_size);    
    scm_array_handle_release(&arrayHandle);
}

void gc(void) {
    SCM gcFunc = scm_variable_ref(scm_c_lookup("gc"));
    scm_call_0(gcFunc);
}

void shutdown(void) {
    /* Didn't find any way to do Guile shutdown. */
}
