/*
Binary of Babel
Copyright (C) 2010 Segfault Garden

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <inttypes.h>
#include <boost/thread.hpp>
#include <iostream>

#include "demolib.h"
#include "pluginthread.h"
#include "demoplugin.h"
#include "demolib.h"

using namespace std;

class PluginCallable {
public:
    PluginCallable(DemoPlugin &plugin, boost::mutex &mutex, 
                   boost::condition_variable &cond, uint32_t &prevFrameTime,
                   uint32_t &nextFrameTime, bool &hasNewFrame,
                   const char *pluginName, int width, int height, const char *configFile)
        : plugin(plugin), mutex(mutex), cond(cond), 
          prevFrameTime(prevFrameTime), nextFrameTime(nextFrameTime), 
          hasNewFrame(hasNewFrame), pluginName(pluginName), width(width), height(height),
          configFile(configFile) {
    }

    void operator()() {
        log_print(debugLevel, "Loading and initializing plugin %s: width=%d, height=%d, configFile=%s\n",
                  pluginName, width, height, configFile);
        plugin.load();
        plugin.init(width, height, configFile);
        try {
            boost::unique_lock<boost::mutex> lock(mutex);
            do {
                while (nextFrameTime == prevFrameTime) {
                    cond.wait(lock);
                }
                
                plugin.render(nextFrameTime);
                hasNewFrame = true;
                prevFrameTime = nextFrameTime;

                cond.notify_one();
            } while (!boost::this_thread::interruption_requested());
        } catch (boost::thread_interrupted e) {
        }
        log_print(traceLevel, "Shutting down plugin.\n");

        plugin.shutdown();
        log_print(traceLevel, "Finished thread.\n");
    }

    // Thanks to copy on thread construction, everything must be references
    uint32_t &prevFrameTime;
    uint32_t &nextFrameTime;
    volatile bool &hasNewFrame;

    DemoPlugin &plugin;
    boost::mutex &mutex;
    boost::condition_variable &cond;

    int width, height;
    const char *configFile;
    const char *pluginName;
};


PluginThread::PluginThread(const char *pluginName, int width, int height, const char *configFile)
    : plugin(pluginName), mutex(), cond(), prevFrameTime(-1), nextFrameTime(-1),
      newFrame(false),
      callable(new PluginCallable(plugin, mutex, cond, prevFrameTime, 
                                  nextFrameTime, newFrame, pluginName, width, height, configFile))
{
    thread = boost::thread(*callable);
}

bool PluginThread::hasNewFrame(void) {
    return callable->hasNewFrame;
}

unsigned char* PluginThread::getTexture() {
    boost::unique_lock<boost::mutex> lock(mutex);
    while (!hasNewFrame())
        cond.wait(lock);

    newFrame = false;
    return plugin.getTexture();
}

void PluginThread::continueToNextFrame(uint32_t time_milliseconds) {
    boost::lock_guard<boost::mutex> lock(mutex);

    assert(prevFrameTime == nextFrameTime);
    /* Trying to ensure that frames are never repeated was annoying... */
    if (time_milliseconds == prevFrameTime) {
        newFrame = true;
        return;
    }

    nextFrameTime = time_milliseconds;
    cond.notify_one();
}

void PluginThread::shutdown(void) {
    log_print(traceLevel, "Shutting down PluginThread.\n");
    thread.interrupt();
    log_print(traceLevel, "Interrupted");
    thread.join();
    log_print(traceLevel, "Joined");
}

PluginThread::~PluginThread(void) {
    delete callable;
}
