/*
Binary of Babel
Copyright (C) 2010 Segfault Garden

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <string>
#include <math.h>
#include "SDL.h"
#include "SDL_image.h"
#include "scenerunner.h"
#include "demorunner.h"
#include "demolib.h"


using namespace std;

const double frameTime = 1000.0/60.0;

SceneRunner::SceneRunner(DemoRunner &dr, VideoManager &vm) : d(dr), v(vm) {
}

/*
 * Simple test function that runs all plugins for a specified time in a loop.
 */

void SceneRunner::slideshowLoop() {
    SDL_Event e;
    Uint32 firstFrameTime;
    PluginThread *plt = d.pluginThreads[0];
    glColor3d(1.0, 1.0, 1.0);
    GLuint texid[1];
    Uint32 fpsCountStart;
    const int TEXSIZE=256;
    int framesRendered = 0;
    int mainIterated = 0;
    int currentPlugin = 0;
    int nFrames = 0;
    int numPlugins = d.pluginThreads.size();
    if (d.textureMap.find(string("testimage")) == d.textureMap.end()) {
        throw "Could not find \"testimage\" in textureMap.\n";
    }
    const GLuint overlayTex = d.textureMap[string("testimage")];

    glGenTextures(1, &texid[0]);
    glBindTexture(GL_TEXTURE_2D, texid[0]);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

    firstFrameTime = fpsCountStart = d.getTime();
    plt->continueToNextFrame(firstFrameTime);

    Uint32 effectStart = firstFrameTime;

    while(1) {
        const Uint32 frameTimeWith60FPS = 17;
        Uint32 startTime = d.getTime();
        Uint32 endTime, elapsed;

        v.clear();
        v.switchTo3DMode();
        glBindTexture(GL_TEXTURE_2D, texid[0]);

        if (plt->hasNewFrame()) {
            glTexImage2D(GL_TEXTURE_2D, 0, 4, TEXSIZE, TEXSIZE, 0, GL_RGBA,
                         GL_UNSIGNED_BYTE, plt->getTexture());
            plt->continueToNextFrame(startTime - firstFrameTime);
        }
        v.simpleDraw(0.0, 0.0);

        v.switchTo2DMode();
        v.drawOverlay(overlayTex, -3.5 * cos((startTime % 4000) / 4000.0 * 2.0
                * M_PI), -2.5 * sin((startTime % 4000) / 4000.0 * 2.0 * M_PI),
                1.0, 1.0, 0.8);

        v.flip();
        framesRendered++;
        if (startTime > effectStart + 2000) {
            currentPlugin = (currentPlugin + 1) % numPlugins;
            log_print(messageLevel, "Switching to plugin %d\n", currentPlugin);
            plt = d.pluginThreads[currentPlugin];
            plt->continueToNextFrame(startTime - firstFrameTime);
            effectStart = startTime;
        }

        // Loop at 60 FPS.
        endTime = d.getTime();
        elapsed = endTime - startTime;
        if (elapsed < frameTimeWith60FPS) {
            SDL_Delay(frameTimeWith60FPS - elapsed);
        }
        mainIterated++;

        // Print FPS rate.
        if (d.getTime() - fpsCountStart > 1000) {
            log_print(messageLevel, "FPS: %d - main iterations: %d\n", framesRendered, mainIterated);
            framesRendered = 0;
            mainIterated = 0;
            fpsCountStart = d.getTime();
        }

        // Check if we need to exit.
        while(SDL_PollEvent(&e)) {
            if (e.type == SDL_QUIT ||
                (e.type == SDL_KEYDOWN &&
                (e.key.keysym.sym == SDLK_ESCAPE || e.key.keysym.sym == SDLK_q))) {
                glDeleteTextures(1, &texid[0]);
                return;
            }
        }
    }
}


void SceneRunner::getTexStuff(vector<string> &plugins, vector<PluginThread*> &threads, vector<GLuint> &texids) {
    PluginThread *plt;
    for (int i=0; i<plugins.size(); i++) {
        string name = plugins[i];
        PluginThread *plt;
        GLuint texid;
        assert(d.pluginThreadNameMap.find(name) != d.pluginThreadNameMap.end());
        assert(d.textureMap.find(name) != d.textureMap.end());
        plt = d.pluginThreadNameMap[name];
        texid = d.textureMap[name];
        threads.push_back(plt);
        texids.push_back(texid);
    }
}


void SceneRunner::delayForFrameFlip(int numFramesRendered, Uint32 startTime, Uint32 frameStart, const char *sceneName) {
    /* This is a no-op, since we are currently using VSYNC to deal with flipping. */
}

void SceneRunner::updateTextures(Uint32 frameStart, vector<PluginThread*> &threads, vector<GLuint> &texids) {
    assert(threads.size() == texids.size());
    for (int i=0; i<threads.size(); i++) {
        PluginThread *plt = threads[i];
        GLuint texid = texids[i];
        if (plt->hasNewFrame()) {
            d.updatePluginTexture(texid, plt);
            plt->continueToNextFrame(frameStart - d.zeroTime);
        }
    }
}

void SceneRunner::runCubeScene(Uint32 startTime, basicScene &s) {
    sceneStartTime = startTime;
    framesRenderedInThisScene = 0;
    runIt(s, cubeScene);
}

void SceneRunner::runIntersectScene(Uint32 startTime, basicScene &s) {
    sceneStartTime = startTime;
    framesRenderedInThisScene = 0;
    runIt(s, isectScene);
}


void SceneRunner::runTestScene(Uint32 startTime, basicScene &s) {
    sceneStartTime = startTime;
    framesRenderedInThisScene = 0;
    runIt(s, testScene);
}

void SceneRunner::runEmptyScene(Uint32 startTime, basicScene &s) {
    sceneStartTime = startTime;
    framesRenderedInThisScene = 0;
    runIt(s, emptyScene);
}

void SceneRunner::runRayScene(Uint32 startTime, basicScene &s) {
    GLuint texid = 0;
    const int FPS = 20;
    const double frameTime = 1000.0/FPS;
    const char *fnameTemplate = "out/world_animation%03d.png";

    v.switchTo2DMode();
    do {
        v.clear();
        int frameNum = (SDL_GetTicks() - startTime) / frameTime + 1;
        char fname[1024];
        snprintf(fname, 1024, fnameTemplate, frameNum);
        SDL_Surface *frame = IMG_Load(fname);
        if (!frame) {
            //log_print(debugLevel, "Could not load image %s.\n", fname);
            SDL_Delay(frameTime);
        } else {
            assert(frame->format->BitsPerPixel == 24);
            if (texid == 0) {
                texid = v.createTexture(frame->w, frame->h, frame->format->BitsPerPixel / 8, true, GL_RGB, (unsigned char*) frame->pixels);
                log_print(traceLevel, "Created raytrace texture: %d\n", texid);
            } else {
                log_print(traceLevel, "Updating: %d\n", texid);
                glBindTexture(GL_TEXTURE_2D, texid);
                glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, frame->w, frame->h, GL_RGB, GL_UNSIGNED_BYTE, frame->pixels);
            }
            v.drawOverlay(texid, 0, 0, 8, 8 * frame->h / frame->w, 1.0);
            v.flip();
            SDL_FreeSurface(frame);
            int sleepTime = (startTime + frameNum * frameTime) - (int) d.getTime();
            if (sleepTime > 0)
                SDL_Delay(sleepTime);
        }
        checkForExit();
    } while (d.getTime() - startTime < s.duration);
    v.deleteTexture(texid);
}


void SceneRunner::renderTestScene(vector<GLuint> &texids, Uint32 timeSinceSceneStart) {
    v.bind(texids[0]);
    v.simpleDraw(0.0, 0.0);
}

void SceneRunner::renderCubeScene(vector<GLuint> &texids, Uint32 timeSinceSceneStart) {
    double alfa, beta;
    alfa = (timeSinceSceneStart % 10000)/10000.0*360.0;
    beta = (timeSinceSceneStart % 3000)/3000.0*360.0;
    glTranslated(0.0, 0.6, -6.5);
    glRotated(alfa, 1.0, 0.0, 0.0);
    glRotated(beta, 0.0, 1.0, 0.0);
    switch (texids.size()) {
    case 1 : v.drawCube(texids[0]); break;
    case 6 : v.drawCube(texids); break;
    default : throw "Unsupported amount of textures for cube.";
    }
}
void SceneRunner::renderIntersectScene(std::vector<GLuint> &texids, Uint32 timeSinceSceneStart) {
    assert(texids.size() >= 2);
    double alfa, beta, alfa2, beta2, shiftx;
    alfa = (timeSinceSceneStart % 10000)/10000.0*360.0;
    beta = (timeSinceSceneStart % 3000)/3000.0*360.0;
    alfa2 = (timeSinceSceneStart % 7000)/7000.0*360.0;
    beta2 = (timeSinceSceneStart % 4000)/4000.0*360.0;
    shiftx = 3*cos(timeSinceSceneStart/5000.0*2.0*M_PI)/exp(timeSinceSceneStart/3250.0);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glTranslated(-shiftx, 0.6, -6.5);
    glRotated(alfa, 1.0, 0.0, 0.0);
    glRotated(beta, 0.0, 1.0, 0.0);
    v.drawCube(texids[0]);
    glPopMatrix();
    glPushMatrix();
    glTranslated(shiftx, 0.6, -6.5);
    glRotated(alfa2, 1.0, 0.0, 0.0);
    glRotated(beta2, 0.0, 1.0, 0.0);
    v.drawCube(texids[1]);
    glPopMatrix();
}

void SceneRunner::runIt(basicScene &s, sceneType type) {
    Uint32 frameStart, frameEnd, elapsedTime;
    vector<PluginThread*> threads;
    vector<GLuint> texids;

    getTexStuff(s.plugins, threads, texids);

    do {
        frameStart = d.getTime();
        v.clear();
        v.switchTo3DMode();
        updateTextures(frameStart, threads, texids);
        switch (type) {
        case testScene : renderTestScene(texids, frameStart - sceneStartTime); break;
        case cubeScene : renderCubeScene(texids, frameStart - sceneStartTime); break;
        case isectScene : renderIntersectScene(texids, frameStart - sceneStartTime); break;
        case emptyScene : break;
        }
        elapsedTime = frameStart - sceneStartTime;
        d.drawOverlays(s.overlays, elapsedTime);
        framesRenderedInThisScene++;
        delayForFrameFlip(framesRenderedInThisScene, sceneStartTime, frameStart, "test");
        v.flip();
        log_print(traceLevel, "Scene test frame at %d.\n", frameStart);
        checkForExit();
    } while (elapsedTime < s.duration);
}


void SceneRunner::checkForExit() {
    SDL_Event e;
    while (SDL_PollEvent(&e)) {
        if (e.type == SDL_QUIT || (e.type == SDL_KEYDOWN && (e.key.keysym.sym
                == SDLK_ESCAPE || e.key.keysym.sym == SDLK_q))) {
            throw new UserExit();
        }
    }
}
