/*
Binary of Babel
Copyright (C) 2010 Segfault Garden

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <string>
#include <assert.h>
#include "videomanager.h"
#include "demolib.h"
#include "config.h"

using namespace std;

VideoManager::VideoManager(int width, int height, bool fullScr) {
    screenWidth = width;
    screenHeight = height;
    isFullscreen = fullScr;
}


void VideoManager::updateTexture(GLuint texid, int width, int height, unsigned char *pixels) {
    glBindTexture(GL_TEXTURE_2D, texid);
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
    if (glGetError() != GL_NO_ERROR) {
        throw "OpenGL texture update failed.";
    }
}

void VideoManager::switchTo2DMode() {
    glDisable(GL_DEPTH_TEST);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(-4.0, 4.0, -3.0, 3.0);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glColor4d(1.0, 1.0, 1.0, 1.0);
}

void VideoManager::switchTo3DMode() {
    glEnable(GL_DEPTH_TEST);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0f,(GLfloat)screenWidth/(GLfloat)screenHeight,0.1f,100.0f);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glColor4d(1.0, 1.0, 1.0, 1.0);
}



void VideoManager::initGraphics() {
    SDL_Surface *canvas;
    int val;
    if (SDL_Init(SDL_INIT_AUDIO|SDL_INIT_VIDEO) < 0) {
        string error = "Unable to init SDL: ";
        error += SDL_GetError();
        throw error;
    }
    SDL_WM_SetCaption(PROGRAM_NAME, PROGRAM_NAME);

    Uint32 surfaceFlags = SDL_OPENGL;
    int depth = 32;
    if (isFullscreen)
        surfaceFlags |= SDL_FULLSCREEN;
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
    SDL_GL_SetAttribute(SDL_GL_SWAP_CONTROL, 1); // Use VSYNC.
    canvas = SDL_SetVideoMode(screenWidth, screenHeight, depth, surfaceFlags);
    if (!canvas) {
        log_print(errorLevel, "Could not initialize video: %s\n", SDL_GetError());
        exit(1);
    }
    initOpenGL();
    if (isFullscreen) {
        SDL_ShowCursor(SDL_DISABLE);
    }

    SDL_GL_GetAttribute(SDL_GL_DOUBLEBUFFER, &val);
    if (val != 1) {
        throw "Could not set OpenGL double buffering.\n";
    }

    SDL_GL_GetAttribute(SDL_GL_SWAP_CONTROL, &val);
    if (val != 1) {
        throw "Can not use VSYNC.\n";
    }

    log_print(debugLevel, "OpenGL vendor:   %s\n", glGetString(GL_VENDOR));
    log_print(debugLevel, "OpenGL renderer: %s\n", glGetString(GL_RENDERER));
    log_print(debugLevel, "OpenGL version:  %s\n", glGetString(GL_VERSION));
    log_print(debugLevel, "Screen size:     %dx%d.\n", canvas->w, canvas->h);
}

GLuint VideoManager::createTexture(int w, int h, int depth, bool smooth, GLuint pixelFormat, unsigned char *pixels) {
    GLuint texid[1];
    glGenTextures(1, &texid[0]);
    glBindTexture(GL_TEXTURE_2D, texid[0]);

    if (glGetError() != GL_NO_ERROR) {
        throw "OpenGL texture creation failed.";
    }

    if (!smooth) {
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    } else {
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    }
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

    glTexImage2D(GL_TEXTURE_2D, 0, depth, w, h, 0, pixelFormat, GL_UNSIGNED_BYTE, pixels);
    if (glGetError() != GL_NO_ERROR) {
        throw "OpenGL texture upload failed.";
    }
    return texid[0];
}


void VideoManager::shutdown() {
    SDL_Quit();
}

void VideoManager::initOpenGL() {
    glShadeModel(GL_SMOOTH);
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClearDepth(1.0f);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

/*
 * (x, y) is the center point of the texture.
 */

void VideoManager::drawOverlay(GLuint tex, double x, double y, double width, double height, double opacity) {
    assert(width > 0);
    assert(height > 0);
    assert(opacity >= 0.0);
    assert(opacity <= 1.0);
    glColor4d(1.0, 1.0, 1.0, opacity);
    glMatrixMode(GL_MODELVIEW);
    glBindTexture(GL_TEXTURE_2D, tex);
    glPushMatrix();
    glTranslated(x, y, 0);
    glScaled(width, height, 0.0);
    glScaled(0.5, 0.5, 1.0);
    drawBasicPlane();
    glPopMatrix();
}

void VideoManager::simpleDraw(double dx, double dy) {
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glTranslated(dx, dy, -2.5f);
    drawBasicPlane();
    glPopMatrix();
}

void VideoManager::drawBasicPlane() {
    glBegin(GL_QUADS);
    glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f, -1.0f,  0.0f);
    glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f, -1.0f,  0.0f);
    glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f,  1.0f,  0.0f);
    glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f,  1.0f,  0.0f);
    glEnd();
}

void VideoManager::drawCube(vector<GLuint> &texes) {
    assert(texes.size() >= 6);
    glMatrixMode(GL_MODELVIEW);

    // Now draw the faces one by one. Start with front.
    glBindTexture(GL_TEXTURE_2D, texes[0]);
    glPushMatrix();
    glTranslated(0.0, 0.0, -1.0);
    glRotated(180, 0.0, 1.0, 0.0);
    drawBasicPlane();
    glPopMatrix();

    // Left.
    glBindTexture(GL_TEXTURE_2D, texes[1]);
    glPushMatrix();
    glTranslated(-1.0, 0.0, 0.0);
    glRotated(-90, 0.0, 1.0, 0.0);
    drawBasicPlane();
    glPopMatrix();

    // Right
    glBindTexture(GL_TEXTURE_2D, texes[2]);
    glPushMatrix();
    glTranslated(1.0, 0.0, 0.0);
    glRotated(90, 0.0, 1.0, 0.0);
    drawBasicPlane();
    glPopMatrix();

    // Back.
    glBindTexture(GL_TEXTURE_2D, texes[3]);
    glPushMatrix();
    glTranslated(0.0, 0.0, 1.0);
    drawBasicPlane();
    glPopMatrix();

    // Top
    glBindTexture(GL_TEXTURE_2D, texes[4]);
    glPushMatrix();
    glTranslated(0.0, 1.0, 0.0);
    glRotated(-90, 1.0, 0.0, 0.0);
    drawBasicPlane();
    glPopMatrix();

    // bottom
    glBindTexture(GL_TEXTURE_2D, texes[5]);
    glPushMatrix();
    glTranslated(0.0, -1.0, 0.0);
    glRotated(90, 1.0, 0.0, 0.0);
    drawBasicPlane();
    glPopMatrix();
}

void VideoManager::drawCube(GLuint tex) {
    vector<GLuint> texes;
    for (int i=0; i<6; i++)
        texes.push_back(tex);
    drawCube(texes);
}
